//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file calrec.c
//!
//! \brief  Calibration record management
//!
//! Handle the non-volatile calibration data
//!
//! \author David R Brooks
//!
//! \version 28/01/2008 DRB initial design and coding
//!
//---------------------------------------------------------------------------


//-Included files -----------------------------------------------------------
#include "main.h"
#include "avr/eeprom.h"
#include "avr/pgmspace.h"

//-Local Symbols ------------------------------------------------------------
// [insert local #defines and variable types]

//-Globals ------------------------------------------------------------------
T_CALIBN CalRam;			// Public calibration data

//-Locals -------------------------------------------------------------------
static T_CALIBN CalEerom[2] eeprom;	// The record in EEPROM	(2 copies)

const T_CALIBN CalRamDefault flash = // Default calibration data
{
	57600,									// Baud-rate
    {15, 30},                               // V1 cathode current (mA*10)
    { 7, 20},                               // V2 cathode current (mA)
    
  	{										// Bogey servo settings
  		{(uint4)286 << ACCUMSHIFT, 16,   0}, 	// RU 
    	{(uint4)800 << ACCUMSHIFT, 16, 250},	// LL
    	{(uint4)800 << ACCUMSHIFT, 16, 250},	// RL
    	{(uint4)286 << ACCUMSHIFT, 16,   0}		// LU
    },
    64,										// Scale factor for quick gain (calibn.)
    {-50, +50},								// OP mV must be within this, at start
    {240, 260},								// OP mA must be within this, at start
	{(uint4)600 << ACCUMSHIFT, 64, 1000}, 	// Clip-voltage driver 
    20,										// State-machine period (mS)
    5000,									// Startup delay (uS)
    { 600,  720},							// Limits for +660
    { 300,  380},							// Limits for +330
    { 120,  165},							// Limits for +140
    {-165, -120},							// Limits for -140
    {-380, -280},							// Limits for -330
    1000,                                   // Max. filtered OP current (mA)
    
	0x5555,									// Invalid sumcheck
};

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

static uint2 CalCheckSum(T_CALIBN *pRec);
static void  eeprom_write(const void *buf, void *addr, size_t n);

//! \endcond

//-Local Functions ----------------------------------------------------------

//--CalCheckSum----------------------------------------------------------------
//! Compute checksum
//-----------------------------------------------------------------------------
//! Checksum over the calibration record (EXCLUDING its own checksum)
//! 
//! \param pRec       Ptr. to cal. record
//! \return           Computed checksum
//! 
//-----------------------------------------------------------------------------
static uint2 CalCheckSum(T_CALIBN *pRec)
{
	uint1 *ptr = (uint1*)pRec;
    int i;
    uint2 sum  = 0;
    
    for(i=0; i<(sizeof(T_CALIBN) - sizeof(pRec->sumcheck)); i++)
    {
    	sum += *(ptr++);
		_WDR();
    }
    return sum;
}

//-Global Functions ---------------------------------------------------------

//--CalibrationDefault---------------------------------------------------------
//! Reset default calibration data
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//-----------------------------------------------------------------------------
void CalibrationDefault(void)
{
	memcpy_P(&CalRam, &CalRamDefault, sizeof(CalRam));
    CalibrationWrite(BOTH);
}

//--CalibrationRead------------------------------------------------------------
//! Read EEROM calibration to RAM
//-----------------------------------------------------------------------------
//! Read the EEROM copy. If the checksum is OK, copy it to RAM.
//! Try the first EEROM copy, then the second. The first good one is accepted.
//! Otherwise, fetch the default from Flash.
//! If one copy is bad, overwrite it with the good one.
//! 
//! \return           TRUE if one EEROM was OK, & was copied
//! 
//-----------------------------------------------------------------------------
bool1 CalibrationRead(void)
{
	T_CALIBN temp[2];		// Temporary copies
    bool1    good[2];
    uint1    i;

	for(i=0; i<2; i++)		// Read both
    {    
    	eeprom_read_block(&temp[i], &CalEerom[i], sizeof(T_CALIBN));
    	good[i] = (temp[i].sumcheck == CalCheckSum(&temp[i]));
    }

	if(good[0])				// 1st is OK?
    {
    	memcpy(&CalRam, &temp[0], sizeof(T_CALIBN));	// Copy to public record
        if(!good[1])
        {						// Copy over bad second
    		CalibrationWrite(SECOND);
        }
    }
	else
	if(good[1])				// 2nd is OK?
    {
    	memcpy(&CalRam, &temp[1], sizeof(T_CALIBN));	// Copy to public record
        if(!good[0])
        {						// Copy over bad first
    		CalibrationWrite(FIRST);
        }
    }
	else
    {						// Neither was good
    	CalibrationDefault();		// Fetch default from Flash
    	return FALSE;		
    }
    return TRUE;
}

//--CalibrationWrite-----------------------------------------------------------
//! Store RAM calibration in EEROM
//-----------------------------------------------------------------------------
//! There are 2 copies (for safety) in EEROM. Write the first, then the second.
//! If power fails while writing the first, the second still holds the old
//! data.
//! 
//-----------------------------------------------------------------------------
void  CalibrationWrite(CalWrite bank)
{
	CalRam.sumcheck = CalCheckSum(&CalRam);	// Make a good checksum
    switch(bank)
    {
    case BOTH:
	    eeprom_write(&CalRam, &CalEerom[0], sizeof(T_CALIBN));
	    eeprom_write(&CalRam, &CalEerom[1], sizeof(T_CALIBN));
        break;
    
    case FIRST:
	    eeprom_write(&CalRam, &CalEerom[0], sizeof(T_CALIBN));
        if(bank == FIRST)
        {
        	break;
        }
    case SECOND:
	    eeprom_write(&CalRam, &CalEerom[1], sizeof(T_CALIBN));
        break;
    }
}

//--eeprom_write---------------------------------------------------------------
//! Write a block to EEPROM
//-----------------------------------------------------------------------------
//! We can't use the library, as it ignores the watchdog
//! 
//! \param buf        RAM (source) ptr.
//! \param addr       EEROM (target) ptr.
//! \param n          length (bytes)
//! 
//-----------------------------------------------------------------------------
static void eeprom_write(const void *buf, void *addr, size_t n)
{
	size_t i;
    uint1 *r = (uint1*)buf;
    uint1 *e = (uint1*)addr;
    
    for(i=0; i<n; i++)
    {
    	while(!	eeprom_is_ready() )
        {
    		_WDR();
        }
        eeprom_write_byte(e++, *(r++));
    }
}

