//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file main.c
//!
//! \project Igor-IIL 25+25W OTL valve amplifier
//!
//! \brief  Valve amplifier control: mainline code
//!
//! Main-line code: supervisory functions
//!     Processor:  Atmel ATMega8
//!     Compiler:   GCC          
//!
//! \author David R Brooks
//!
//! \version 14/07/2007 DRB initial design and coding
//!          25/08/2007     fixed bugs. Stable release
//!
//---------------------------------------------------------------------------
// 2008.08.19 18:42  [YYYY.MM.DD] Last edited (CodeWright)

/* Abstract:
  The system comprises the following autonomous modules:
  1. Main state machine: driven from T2, at 20mS (nominal)
  2. AC phase generator: driven from the AC line, via the Analogue Comparator
     Uses T0 to generate the delayed triggers for the SSRs.
  3. SPI output: runs at IO speed (clock = 62.5kHz). Outputs the 4 DAC
     values (OP tube bias) & the '595 output register.
  4. ADC input: runs at IO speed (104uS per conversion). All ADC inputs are
     heavily filtered, so timing is not critical.
  5. UART driver: a basic queued output. Input is un-buffered (keyboard).
  6. INT0 is a fast IO-error signal: eg output overload.

  The autonomous IO drivers maintain RAM buffers: the main code need only 
  write new data to the buffer, & it will be automatically output on the
  next cycle.
  
  Internally, physical units are normally V, mV, mA, Seconds.
*/  

//-Included files -----------------------------------------------------------
#include "main.h"
#include <avr/pgmspace.h>

//-Local Symbols ------------------------------------------------------------
#define BLKSLOW	    	 25		// Slow LED blink-time
#define BLKFAST	    	 12		// Fast ditto

#define LTDEL       	100		// Pause (state periods) between each LT phase advance
#define HTDEL       	 30		// Pause (state periods) between each HT phase advance
#define PHSTEP      	 25		// Phase-advance step
#define MINPHS      	 75		// Min. adjustable phase
#define MAXPHS      	150		// Max. adjustable phase
#define WARMSEC     	 20		// Seconds at full LT before HT
#define DEBOUNCETIME	  5		// Switch debounce time, in state periods

typedef struct                  // HT voltage-check structure
{
    uint2 *sml;                 // sml < big to pass
    uint2 *big;
    uint1  msg[7];              // Message if it fails
} T_CHK;

//-Globals ------------------------------------------------------------------
const uint1 Fill[] flash = "  ";
 
bool1    faultdet  = FALSE;		// Fault detected
T_SWITCH panelsw   = SW_LOCOFF;	// Panel/remote switch settings
T_STATE  state     = M_OFF;		// Current machine state
uint2    seconds   = 0;			// Time since start-up
PGM_P    FaultMsg  = Fill;		// Fault-message ptr. (ROM)
PGM_P    SwitchMsg = Fill;		// Panel-switch state
bool1    HalfSec   = FALSE;		// Half-second prescaler
T_MAINST cstate;				// Current state. Also visible to Logger
uint2    IV1[2];                // V1 cathode current	

//-Locals -------------------------------------------------------------------
static uint2 delay;				// State machine timer
static uint1 phase;             // SSR firing angle
static bool1 LogMode   = TRUE;  // Log mode switch
static bool1 timerflag = FALSE;	// Run the state machine

static uint1 ledslow   = 0;		// Panel LED timers
static uint1 ledfast   = 0;
static uint1 slowflag  = BV(LEDHT);
static uint1 fastflag  = BV(LEDHT);

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

static void  ShutDown(void);
static void  KeyActions(void);
static void  ShowChoices(void);
static void  StateMachine(void);
static void  GetSwitch(void);
static void  IoPortInit(void);
static void  LedTimer(uint1 *pTimer, uint1 *pFlag, const uint1 Limit);
static uint1 DoLed(LedState state, uint1 bit);
static bool1 CheckHT(CKVAL span);
static uint2 GetIV1(uint2 Ik2);

//! \endcond

//-Local Functions ----------------------------------------------------------

//--IoPortInit-----------------------------------------------------------------
//! Set up pin directions, etc.
//-----------------------------------------------------------------------------
//! Set up the default pin values & directions
//! 
//! \note             Reset defines all port & DDR regs. as zero (input)
//! 
//-----------------------------------------------------------------------------
static void IoPortInit(void)
{
    PORTB = REMSTBY | REMRUN;		// Activate pullups
    DDRB  = INIT_DDR_B;
    PORTC = 0;
    DDRC  = INIT_DDR_C;
    PORTD = BV(PD3) | BV(PD5);		// Active-low chip enables
    DDRD  = INIT_DDR_D;
}

//--ShutDown-------------------------------------------------------------------
//! Shut down the amplifier
//-----------------------------------------------------------------------------
//!
//-----------------------------------------------------------------------------
static void ShutDown(void)
{
    delay    = 0;
	Reg595   = (BV(KSWL) | BV(KSWR));	// Turn everything off
    faultdet = FALSE;			        // Allows the 74HC595 to load
    SetPhase(0, 0);				        // SSR phase drives off
}
    
//--KeyActions-----------------------------------------------------------------
//! Interpret user keystrokes (special commands)
//-----------------------------------------------------------------------------
//! When a maintenance computer is connected, the user has additional controls,
//! as implemented here.
//! 
//-----------------------------------------------------------------------------
static void KeyActions(void)
{
	static const sint1 movetab[] = {-ITEMSPERROW, ITEMSPERROW, -1, +1, 100};
    static const sint2 edittab[] = {-1, +1, -10, +10, -100, 0, +100, -1000, +1000, -10000, +10000};
    uint1 key = SerialGetRx();		// If none, returns '\0'
    
    switch(key)
    {
    case 'D':						// Default EEROM data
    	CalibrationDefault();
        break;
        
    case 'E':						// Commit EEROM data
    	CalibrationWrite(BOTH);
        break;
        
    case KUP ... KHOME:				// Move cursor around
    	EditMove(movetab[key-KUP]);
        break;
        
    case KF1 ... KF10:				// Change selected cell
    	EditNum(edittab[key-KF1]);
        break;
        
    case KF12:                      // Logging mode
        LogMode = !LogMode;
        break;
        
    default:						// Irrelevant
    	;
    }
}

//--ShowChoices----------------------------------------------------------------
//! Display list of keyboard commands
//-----------------------------------------------------------------------------
//! 
//-----------------------------------------------------------------------------
static void ShowChoices(void)
{
	SetCursor(GetLogSpace()+LOGROW+1, 1);	// Position after the log display
    
    outs_P(PSTR(" Keys:"
		"\n  D                Reset & commit default caln. data"
		"\n  E                Commit caln. data to EEROM"
        "\n  <  >  ^  V       Move cursor on display"
        "\n  F1 F3 F5 F7 F7   Add 1 10 100 1000 10000 to value" 
        "\n  F2 F4 F6 F8 F10  Sub 1 10 100 1000 10000 from value" 
 		));
}

//-Global Functions ---------------------------------------------------------

//--INTERRUPT 0 (fault)--------------------------------------------------------
//! Falling edge = output line out of range
//-----------------------------------------------------------------------------
//! Overvoltage on an output line indicates either a bias failure or a short
//! (probably) in an output valve. Once the HT supply falls, the fault is
//! expected to clear, so the software can re-assert the output register
//! (74HC595), and drive the panel LEDs.
//! 
//-----------------------------------------------------------------------------
SIGNAL(SIG_INTERRUPT0)
{					  
	Reg595   = (BV(KSWL) | BV(KSWR));	// So the automatic SPI driver doesn't restart the SSRs
    faultdet = TRUE;		            // Trap the main state machine
    FaultMsg = PSTR("OP chk");
}

//--SIGNAL(SIG_OUTPUT_COMPARE2)------------------------------------------------
//! Main timebase (T2) interrupt
//-----------------------------------------------------------------------------
//! Main timer interrupt. Run the state machine
//! 
//-----------------------------------------------------------------------------
SIGNAL(SIG_OUTPUT_COMPARE2)
{
    timerflag = TRUE;
}

//--GetSwitch------------------------------------------------------------------
//! Read the control switches
//-----------------------------------------------------------------------------
//! The front-panel switch dominates. If it is set to REMOTE, the remote
//! (current-loop) input is read. Result goes in a global variable.
//!
//! Outputs:	panelsw	current (debounced) switch state
//! 
//-----------------------------------------------------------------------------
static void GetSwitch(void)
{
    static uint1    debounce = 0;
    static T_SWITCH oldsw    = SW_LOCOFF;
		   T_SWITCH newsw;
		   uint1    swbits   = 0;			// State of front-panel switch
    
    if(!ScaledAdc_s2[SW1]) swbits |= 1;		// Switch lines are active low
    if(!ScaledAdc_s2[SW2]) swbits |= 2;
    if(!ScaledAdc_s2[SW3]) swbits |= 4;
    
    switch(swbits)
    {
    case 0:						// Local Off
    	newsw = SW_LOCOFF;
    	break;
    case 1:						// Remote
    	switch(PINB & (REMSTBY | REMRUN))
        {			    		// NB Remotes are active low
        case REMSTBY:			// Stby is high (inactive), Run is active
        	newsw = SW_RUN;
        	break;
        case REMRUN:			// Run is high (inactive), Stby is active
        	newsw = SW_STANDBY;
        	break;
        case REMSTBY | REMRUN:	// Both inactive: remote Off
        	newsw = SW_REMOFF;
            break;
        default:				// Illegal states: ignore
        	return;
        }
    	break;
    case 2:						// Standby
    	newsw = SW_STANDBY;
    	break;
    case 4:						// Run
    	newsw = SW_RUN;
    	break;
    default:					// Transient states
    	return;
    }
    						// Post the result (which is valid)
	if(oldsw == newsw)			// Debounce: N equal values
    {
    	if(debounce == 0)
        {						// Debounce exhausted: post result
    		panelsw = newsw;
            switch(panelsw)
            {
			case SW_LOCOFF:
	    		FaultMsg  = Fill;			// Clear the fault display
            	SwitchMsg = PSTR("LocOff");
                break;
			case SW_REMOFF:
            	SwitchMsg = PSTR("RemOff");
                break;
    		case SW_STANDBY:
            	SwitchMsg = PSTR("Stndby");
                break;
    		case SW_RUN:
            	SwitchMsg = PSTR("   Run");
                break;
            default:
            	SwitchMsg = Fill;
                break;
            }
        }
        else
        {
        	debounce--;
        }
    }
    else
    {
    	debounce = DEBOUNCETIME;
    	oldsw    = newsw;
    }
}

//--DoLed----------------------------------------------------------------------
//! Maintain the panel LEDs
//-----------------------------------------------------------------------------
//! Called from the main state machine
//! 
//! \param state      Current state
//! \param bit        Bit-value for LED
//! 
//-----------------------------------------------------------------------------
static uint1 DoLed(LedState state, uint1 bit)
{    
    switch(state)
    {
    case L_OFF:
    	return 0;
    case L_SLOW:
    	return bit & slowflag;
    case L_FAST:
    	return bit & fastflag;
    case L_ON:
    	return bit;
    }
    return 0;			// Shouldn't happen
}

//--LedTimer-------------------------------------------------------------------
//! Run the fast/slow LED timer
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param pTimer     Ptr. to timer
//! \param pFlag      Ptr. to flags
//! \param Limit      Timer modulus
//! 
//-----------------------------------------------------------------------------
static void LedTimer(uint1 *pTimer, uint1 *pFlag, const uint1 Limit)
{
    if(*pTimer >= Limit)
    {
    	*pTimer  = 0;
        *pFlag ^= (BV(LEDHT) | BV(LEDLT));
    }
    else
    {
    	(*pTimer)++;
    }
}
    
//--ManageLEDs-----------------------------------------------------------------
//! Manage the panel LEDs
//-----------------------------------------------------------------------------
//! Manage the panel LEDs (actions determined by current state)    
//! 
//-----------------------------------------------------------------------------
static void ManageLEDs(void)
{
    uint1  treg = 0;

    LedTimer(&ledslow, &slowflag, BLKSLOW);
    LedTimer(&ledfast, &fastflag, BLKFAST);
    
    treg |= DoLed(cstate.green, BV(LEDHT));
    treg |= DoLed(cstate.red,   BV(LEDLT));
    
    Reg595 = (Reg595 & ~(BV(LEDHT) | BV(LEDLT))) | treg;
}

//--Prescaler-----------------------------------------------------------------
//! Manage the elapsed-time counter/s
//-----------------------------------------------------------------------------
//! Manage the elapsed-time counter/s
//! 
//-----------------------------------------------------------------------------
static void Prescaler(void)
{
    static uint1 prescale = 0;

    if(prescale == 0)					// Seconds from start, stop at 65000 = 18hrs.
    {
    	prescale = 500 / CalRam.period_ms;
        HalfSec = !HalfSec;				// Half-second prescaler
        
        if(HalfSec && (seconds < 65000))
        {					
        	seconds++;
        }	   
    }
    else
    {
    	prescale--;
    }
}    

///////////////////////////////////////////////////////////////////////////////
//     S T A T E  -  M A C H I N E   M O D U L E                             //
///////////////////////////////////////////////////////////////////////////////

// Slave helper functions
//--sRMP-----------------------------------------------------------------------
//! Ramp-up the HT/LT supply
//-----------------------------------------------------------------------------
//! Ramp-up the HT/LT supply
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sRMP(bool1 new, uint2 del)
{
    if(new)
    {
    	phase = MINPHS - PHSTEP;
        delay = 0;
    }

	if(delay > 0)
    {
    	delay--;				// Time delay
		return FALSE;
    }
    
    if(phase < MAXPHS)
    {
    	phase += PHSTEP;		// Advance the LT phase
        delay  = del;
		return FALSE;
    }
    
	return TRUE;				// Fully advanced to Standby setting
}

//--InCal----------------------------------------------------------------------
//! Check OP settings are within calibration (NOT fault) limits
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param current    Observed current
//! \param voltage    Observed voltage
//! \return           TRUE if OK
//! 
//-----------------------------------------------------------------------------
static bool1 InCal(sint2 current, sint2 voltage)
{
	return (CalRam.calamp.minv  <= current) && (current <= CalRam.calamp.maxv) &&
    	   (CalRam.calvolt.minv <= voltage) && (voltage <= CalRam.calvolt.maxv);
}

//--sNUL-----------------------------------------------------------------------
//! Do nothing
//-----------------------------------------------------------------------------
//! No special action: next state at once
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sNUL(bool1 new)
{
    return TRUE;
}

//--sFault---------------------------------------------------------------------
//! Action on hardware faults
//-----------------------------------------------------------------------------
//! Fault state: waits for local OFF
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sFLT(bool1 new)
{
    ShutDown();
    return TRUE;
}

//--sOff-----------------------------------------------------------------------
//! Action while switched OFF
//-----------------------------------------------------------------------------
//! Wait in the OFF state
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sOFF(bool1 new)
{
	ShutDown();
	seconds = 0;				// Time from ON
    return TRUE;
}

//--sWRM-----------------------------------------------------------------------
//! Pause while the tubes warm up
//-----------------------------------------------------------------------------
//! Delay to warm-up the cathodes
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sWRM(bool1 new)
{
    if(new)
    {
        delay = WARMSEC * 1000 / CalRam.period_ms;
    }    
    return((delay--) == 0);
}

//--sLTR-----------------------------------------------------------------------
//! Ramp-up the LT supply
//-----------------------------------------------------------------------------
//! Ramp-up the LT supply
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sLTR(bool1 new)
{
    return sRMP(new, LTDEL);
}

//--sHTR-----------------------------------------------------------------------
//! Ramp-up the HT supply
//-----------------------------------------------------------------------------
//! Ramp-up the HT supply
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sHTR(bool1 new)
{
    return sRMP(new, HTDEL);
}

//--sCalib---------------------------------------------------------------------
//! Fine-tune the OP tube bias, before starting
//-----------------------------------------------------------------------------
//! Calibrate the bias generators
//! TODO: We might save the calibration record here, once it's done.
//!       This may need a delay, to match the filter bandwidths
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 sCAL(bool1 new)
{
	if(InCal(ScaledAdc_s2[IDIRL], ScaledAdc_s2[VOUTL]) &&
       InCal(ScaledAdc_s2[IDIRL], ScaledAdc_s2[VOUTL]))
    {									    // Once we're in calibration, advance
        CalibrationWrite(BOTH);			    // Save corrected settings
      	return TRUE;
    }
	return FALSE;
}

//--s100-----------------------------------------------------------------------
//! Wait for 100mS
//-----------------------------------------------------------------------------
//! 
//! \return           Next-state selector
//! 
//-----------------------------------------------------------------------------
static bool1 s100(bool1 new)
{
    if(new)
        delay = 5;      // 100mS
        
    return((delay--) == 0);
}

//-----------------------------------------------------------------------------
// The main state-flow table
//-----------------------------------------------------------------------------
const T_MAINST SmTable[] flash = {		   //      <--- Next State (Panel Switch) --->              <- LEDs ->
//   LT      HT      V2     Spkr   Func. Checker   OffL    OffR     Standby   Run        PID       Red     Green    Name
	{P_OFF,  P_OFF,  S_OFF, S_OFF, sFLT, CK_OUT,  {M_OFF,  M_FAULT, M_FAULT,  M_FAULT }, G_OFF,    L_FAST, L_FAST, "  Fault"}, 
	{P_OFF,  P_OFF,  S_OFF, S_OFF, sOFF, CK_OUT,  {M_OFF,  M_OFF,   M_LTRAMP, M_LTRAMP}, G_OFF,    L_OFF,  L_OFF,  "    Off"}, 
	{P_RAMP, P_OFF,  S_OFF, S_OFF, sLTR, CK_OUT,  {M_OFF,  M_OFF,   M_WARMUP, M_WARMUP}, G_OFF,    L_SLOW, L_OFF,  " LTramp"}, 
	{P_ON,   P_OFF,  S_OFF, S_OFF, sWRM, CK_OUT,  {M_OFF,  M_OFF,   M_STNDBY, M_STNDBY}, G_OFF,    L_SLOW, L_OFF,  " Warmup"}, 
	{P_ON,   P_OFF,  S_OFF, S_OFF, sNUL, CK_OUT,  {M_OFF,  M_OFF,   M_STNDBY, M_HTRAMP}, G_OFF,    L_ON,   L_OFF,  "Standby"}, 
	{P_ON,   P_RAMP, S_OFF, S_OFF, sHTR, CK_NONE, {M_OFF,  M_OFF,   M_STNDBY, M_HTSTDY}, G_OFF,    L_ON,   L_SLOW, " HTramp"}, 
	{P_ON,   P_ON,   S_OFF, S_OFF, s100, CK_HIGH, {M_OFF,  M_OFF,   M_STNDBY, M_CALIB }, G_OFF,    L_ON,   L_SLOW, "HTstedy"}, 
	{P_ON,   P_ON,   S_OFF, S_OFF, sCAL, CK_MOST, {M_OFF,  M_OFF,   M_STNDBY, M_MUTE  }, G_WIDE,   L_FAST, L_ON,   "Calibn."}, 
	{P_ON,   P_ON,   S_OFF, S_OFF, s100, CK_MOST, {M_OFF,  M_OFF,   M_STNDBY, M_V2ON  }, G_NARROW, L_SLOW, L_ON,   "   Mute"}, 
	{P_ON,   P_ON,   S_ON,  S_OFF, s100, CK_MOST, {M_OFF,  M_OFF,   M_STNDBY, M_LIVE  }, G_NARROW, L_SLOW, L_ON,   "   V2On"}, 
	{P_ON,   P_ON,   S_ON,  S_OFF, s100, CK_MOST, {M_V2ON, M_V2ON,  M_V2ON,   M_RUN   }, G_NARROW, L_SLOW, L_ON,   "   Live"}, 
	{P_ON,   P_ON,   S_ON,  S_ON,  sNUL, CK_ALL,  {M_LIVE, M_LIVE,  M_LIVE,   M_RUN   }, G_NARROW, L_ON,   L_ON,   "    Run"}, 
};

//--StateMachine---------------------------------------------------------------
//! The main state machine
//-----------------------------------------------------------------------------
//! State machine runs every 20mS (regardless of AC line frequency)
//! Updates control outputs, & manages state flow.
//! If the slave fn. returns FALSE, control sticks in current state.
//! Fault conditions pre-empt the normal flow.
//! 
//-----------------------------------------------------------------------------
static void StateMachine(void)
{
    static bool1 first = TRUE;
    T_STATE next;
    static uint2 stcnt = 0;
    
    stcnt++;
    
	// Priority tests: these may overrule the state set by the previous
    // invocation.
	if(faultdet || !CheckHT(cstate.LoBound))	// Hardware faults: shut down NOW
    {
		state = M_FAULT;
    }
    else if(!IsAcLineReady())	                // After power-up, measure the AC line freq.
    {							                // Here at power-up, will define cstate
		state = M_OFF;
    }
    										    // Get current-state record
    memcpy_P(&cstate, &SmTable[state], sizeof(cstate));
    
    if(!cstate.statefn(first))                  // Run the state function
        cstate.succ[SW_RUN] = state;            // Override successor if false
        
    next  = cstate.succ[panelsw];               // Successor state
    first = (next != state);
    state = next;
    
    switch(cstate.LT)                           // Power relay: LT (dominant)
    {
    case P_OFF:
    	SetPhase(0, 0);
        break;
    case P_RAMP:
    	SetPhase(0, phase);
        break;
    case P_ON:
        switch(cstate.HT)                       // Power relay: HT (recessive)
        {
        case P_OFF:
    	    SetPhase(0, 255);
            break;
        case P_RAMP:
    	    SetPhase(phase, 255);
            break;
        case P_ON:
    	    SetPhase(255, 255);
            break;
        }
        break;
    }

    if(cstate.V2)                               // NB It's active low from the CPU
    	Reg595 &= ~(BV(KSWL) | BV(KSWR));
    else
    	Reg595 |=  (BV(KSWL) | BV(KSWR));
        
    if(cstate.Spkr)
    	Reg595 |=  BV(SPKREN);
    else
    	Reg595 &= ~BV(SPKREN);

	DacUpdate(cstate.pid);
    ManageLEDs();
}

//--InFault--------------------------------------------------------------------
//! Report if we are in Fault state (so don't update Logs)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \return           TRUE if in Fault
//! 
//-----------------------------------------------------------------------------
bool1 InFault(void)
{
	return (state == M_FAULT);
}

//--CheckHT--------------------------------------------------------------------
//! If reqd., check the HT voltages
//-----------------------------------------------------------------------------
//! Generally, overvoltages are always checked (when the HT is live).
//! Undervoltage is not checked during the HT ramp-up.
//! .
//! Don't (yet) check current minima: it fails unless all valves are fitted.
//! 
//! \param span       What to test
//! \return           Test OK
//! 
//-----------------------------------------------------------------------------
// #define IMINS 1  // Define to check current minima

#ifdef IMINS
#  define CKOF 4
#else
#  define CKOF 0
#endif

const T_CHK TesTab[] flash =    // List of tests to be made 
{                                   // Lower limits: all [0]
#ifdef IMINS
    {&CalRam.iv2[LOW],     &ScaledAdc_s2[IV2L],  "IV2LMn" }, 
    {&CalRam.iv2[LOW],     &ScaledAdc_s2[IV2R],  "IV2RMn" },
    {&CalRam.iv1_10[LOW],  &IV1[LEFT],           "IV1LMn" }, 
    {&CalRam.iv1_10[LOW],  &IV1[RIGHT],          "IV1RMn" }, 
#endif 
                                    // Lower limits: except IV2 [4]
    {&CalRam.vlp660.minv,  &ScaledAdc_s2[P660],  "+660Mn" },
    {&CalRam.vlp330.minv,  &ScaledAdc_s2[P330],  "+330Mn" },
    {&CalRam.vlp140.minv,  &ScaledAdc_s2[P140],  "+140Mn" },
    {&ScaledAdc_s2[M140],  &CalRam.vlm140.maxv,  "-140Mx" },     // -ve rail: reverse tests
    {&ScaledAdc_s2[M330],  &CalRam.vlm330.maxv,  "-330Mx" },
                                    // Upper limits: most times [9]
    {&ScaledAdc_s2[P660],  &CalRam.vlp660.maxv,  "+660Mx" },
    {&ScaledAdc_s2[P330],  &CalRam.vlp330.maxv,  "+330Mx" },
    {&ScaledAdc_s2[P140],  &CalRam.vlp140.maxv,  "+140Mx" },
    {&CalRam.vlm140.minv,  &ScaledAdc_s2[M140],  "-140Mn" },
    {&CalRam.vlm330.minv,  &ScaledAdc_s2[M330],  "-330Mn" },
    {&IV1[LEFT],           &CalRam.iv1_10[HIGH], "IV1LMx" }, 
    {&ScaledAdc_s2[IV2L],  &CalRam.iv2[HIGH],    "IV2LMx" }, 
    {&IV1[RIGHT],          &CalRam.iv1_10[HIGH], "IV1RMx" }, 
    {&ScaledAdc_s2[IV2R],  &CalRam.iv2[HIGH],    "IV2RMx" }, 
                                    // Outputs: always [18]
    {&ScaledAdc_s2[IDIRL], &CalRam.IoutMax,      "Idir_L" },
    {&ScaledAdc_s2[IDIRR], &CalRam.IoutMax,      "Idir_R" },
};    

static bool1 CheckHT(CKVAL span)
{
    uint1 ctr;

    switch(span)
    {
    case CK_ALL:                // Check everything
        ctr = 0;
        break;
    case CK_MOST:               // All except low IV2 (not on yet)
        ctr = CKOF;
        break;
    case CK_HIGH:               // Check high values only
        ctr = 5+CKOF;
        break;
    case CK_OUT:                // Check outputs only
        ctr = 14+CKOF;
        break;
    default:                    // No tests
        return TRUE;
    }
    
    for(; ctr<12; ctr++)
    {
        sint2 a, b;
        a = *(sint2*)pgm_read_word_near(&TesTab[ctr].sml);
        b = *(sint2*)pgm_read_word_near(&TesTab[ctr].big);
        
        if(a > b)
        {                       // Test fails
            FaultMsg = &TesTab[ctr].msg;
            return FALSE;
        }
    }
	return TRUE;		        // All tests succeed, or suppressed
}

//--ScreenSetup----------------------------------------------------------------
//! Initialise display
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param calok      Caln. RAM status
//! 
//-----------------------------------------------------------------------------
void ScreenSetup(bool1 calok)
{
    ClearScreen();                  // Initial sign-on to debug screen
    outs_P(PSTR(SIGNON));           // Sign-on message, (in program flash)
    _WDR();
    ShowChoices();
    _WDR();
    outs_P(PSTR("\nEEROM "));
    outs_P(calok ? PSTR("OK") : PSTR("Fail"));  
	LogInit();
    _WDR();
}

//--SetWatchdog----------------------------------------------------------------
//! Start the watchdog
//-----------------------------------------------------------------------------
//! 
//-----------------------------------------------------------------------------
void SetWatchdog(void)
{
	return;
    cli;                            // Arm the internal watchdog
    _WDR();
    WDTCR |= BV(WDCE) | BV(WDE);    // Watchdog safety
    WDTCR  = BV(WDE)  | 2;			// Armed, timeout = 65mS
    sei;
}

//--GetIV1---------------------------------------------------------------------
//! Deduce V1 anode current from circuit data
//-----------------------------------------------------------------------------
//! Given the (measured) cathode current in V2, & knowledge of the circuit,
//! determine the cathode current in V1.
//! Units: mA, V, kR
//! 
//! \param Ik2        V2 cathode current (mA)
//! \return           V1 cathode current (mA*10)
//! 
//-----------------------------------------------------------------------------
#define R4  100
#define Rdv  25     // Potential divider, * 10
#define R7   20
#define Vgk  10

static uint2 GetIV1(uint2 Ik2)
{
    uint2 Vg2 = (Ik2 * R7) - Vgk;                           // V2 grid voltage
    uint2 Vt  = ((ScaledAdc_s2[P660] + Vg2) * 10) / Rdv;    // Node voltage, R4,R5,R29
    return ((Vt - Vg2) * 10) / R4;
}

//--main-----------------------------------------------------------------------
//! Mainline function
//-----------------------------------------------------------------------------
//! Initialise everything, then loop, displaying log data. When the timer
//! sets its flag, run the main state machine.
//! 
//-----------------------------------------------------------------------------
int main(void)
{
	bool1 calok;
    
    calok = CalibrationRead();		// These come first
    IoPortInit();
    SetWatchdog();

	// Initialise all hardware resources, in the correct order
    TimebaseInit(CalRam.period_ms);
    CommsInit(CalRam.baud);
    AdcInit();
	AnCompInit();
    SpiInit();
    PwmInit();
    wait(CalRam.startdelay_us);     // Let the comms. start up
	ScreenSetup(calok);
    outs_P(PSTR(" Startup."));
        
	for(;;)
    {
        if(timerflag)           // Time to run the main state machine?
        {            
            timerflag = FALSE;
            _WDR();                 // Reset the internal watchdog
            GetSwitch();			// Read panel switches
            Prescaler();            // Seconds counter
            StateMachine();         // Main state engine
            ClipUpdate();			// Adjust the clip voltage
    		AdcStartCycle(TRUE);	// Then run the ADCs again
            IV1[LEFT]  = GetIV1(ScaledAdc_s2[IV2L]);    // Solve for Ik(V1)
            IV1[RIGHT] = GetIV1(ScaledAdc_s2[IV2R]);
        }
        else					// Secondary functions
        {
         	if(IsTxqEmpty())		// Display next log element
			{						//  but don't race the OP queue
                if(LogMode)
                {
				    ShowNextLogItem();
                }
                else
                {
                    RunningLog();
                }                  		
			}
        	KeyActions();			// User keystrokes			
        }
    }            
    return 0;                   // Never reached
}
