//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file main.h
//!
//! \project Igor-IIL 25+25W OTL valve amplifier
//!
//! \brief  Publicly accessible functions and types for main.c
//!
//! General definitions
//!     Version:    2.01         
//!     Processor:  Atmel ATMega8
//!     Compiler:   GCC          
//!
//! \author David R Brooks
//! \date 14/07/2007
//!
//---------------------------------------------------------------------------
// 2008.08.19 18:25  [YYYY.MM.DD] Last edited (CodeWright)

#ifndef __MAIN_H
#define __MAIN_H

/*
    Security/Configuration Bits:
    Tick the following bits in PonyProg's window:
        [SPIEN]
        BOOTSZ1
        BOOTSZ0
        SUT0
        CKSEL3
        CKSEL1
        CKSEL0
        
 See drivers.h for CPU pin assignments.
 
 Operating principles:
   The ADC is run as fast as possible, using an ADC-driven interrupt function,
 to read the fast inputs. These are compared with limits, then downsampled.
 
   The SPI bus feeds the 74HC595 output register, and the TLV5604 DAC (the real-
 time clock and EEROM are not used, although shown on the schematic). The SPI 
 driver runs autonomously, continuously updating these two devices. Except for 
 the main solid-state relays (SSR), these outputs simply mirror RAM-resident
 buffers. So the core software simply updates these core buffers, and the SPI 
 mirrors these settings to the outputs.
   Besides the buffer register, the two SSR bits have a Clear register. After
 the SPI has sent the buffer values, they are ANDed with the clear bits, so
 potentially resetting them before the next SPI cycle. So the SSR driver can set
 3 states:
 1. Buffer =    Clear = 0  The SSR-drive output is permanently off
 2. Buffer =    Clear = 1  The output is permanently on
 3. Buffer = 1, Clear = 0  A single, short pulse is output on the next SPI cycle
 
   The analogue comparator is triggered on both AC-line crossings. It is used to 
 generate phase-control signals for the solid-state relays feeding the main HT &
 LT transformers. Only one SSR can be phase-controlled at one time. When it is,
 the AC-line transition fires TC0 as a monostable. At the trigger point, TC0 inter-
 rupts, and a short pulse is generated, using the SPI mechanism described above.
   The SSR control operates autonomously, scanning global variables which are set
 by the main state machine.
 
   Use of hardware timers:
   T0 (8 bit):  Phase control for SSRs (monostable from AC crossing)
   T1 (16 bit): Clip voltage PWM output
   T2 (8 bit):  Main state-machine timebase   
*/

//-Included files ------------------------------------------------------------
#include "typedefn.h"
#include "interrupt.h"
#include "string.h"
#include "calrec.h"
#include "adc.h"
#include "phasing.h"
#include "timers.h"
#include "logger.h"
#include "serial.h"
#include "spi.h"

//-Symbols -------------------------------------------------------------------
#define PROJECT     "IgorIIL"
#define DESCRIPTION "Amplifier Control"
#define REVISION    "3.05"
#define AUTHOR      "David R Brooks"
#define YEAR        "2008"

#define SETBIT(port, bit, val)  {if(val) port |= BV(bit); else port &= ~BV(bit);}
#define BITHI(port, bit) {port |=  BV(bit);}
#define BITLO(port, bit) {port &= ~BV(bit);}

#define CLKSCALE             8      // Clock prescale: Fclk = 8kHz

//------------------------------------------------------------------------- 
// Input/Output lines
//-------------------------------------------------------------------------
#define SELRTC(x)       SETBIT(PORTB, PB0, x)   // PB0: OUT, RTC's SPI select line (active high)
                                                // PB1: PWM, OC1A, CLIP level
#define SELLATCH(x)     SETBIT(PORTB, PB2, x)   // PB2: OUT, 74HC595 register latch
                                                // PB3: MOSI
                                                // PB4: MISO
                                                // PB5: SCLK
#define REMSTBY         BV(PB6)                 // PB6: IN,  Remote A (standby)
#define REMRUN          BV(PB7)                 // PB7: IN,  Remote B (run)

                                                // Analogue inputs: Vref = 4.096V
                                                // External mux:    0       1       2      3
                                                // PC0: AN0, mux. IV2_L   ICLIP_R IDIR_R IV2_R
                                                // PC1: AN1, mux.  SW1     SW2     SW3   VCLIP
                                                // PC2: AN2, mux. -330M   +660M   -140M  +140M
                                                // PC3: AN3, mux. ICLIP_L IDIR_R  TEMP   +330M
#define ASEL0(x)        SETBIT(PORTC, PC4, x)   // PC4: OUT, analogue mux. sel. 0
#define ASEL1(x)        SETBIT(PORTC, PC5, x)   // PC5: OUT, analogue mux. sel. 1
                                                // PC6: RST
                                                
                                                // AD6: AN6, Output level, Left
                                                // AD7: AN7, Output level, Right

                                                // PD0: RXD
                                                // PD1: TXD
#define FAULT           (PIND & BV(PD2))        // PD2: IN,  INT0 (/FAULT)
#define DACSET(x)       SETBIT(PORTD, PD3, x)   // PD3: OUT, DAC latch enable
                                                // PD4: unused
#define SELFLASH(x)     SETBIT(PORTD, PD5, x)   // PD5: OUT, Flash ROM enable
                                                // PD6: Analog comparator (mains)
                                                // PD7: Analog comparator (mains)


#define START_ADC(x) {ADMUX = (x & 7); ADCSRA |= BV(ADSC);}
#define EXMUX_ADC(x) {ASEL0((x) & 1);  ASEL1((x) & 2);}

// SPI-latch output bits
#define LEDLT   0           // LT LED (red)
#define LEDHT   1           // HT LED (green)
#define SSRLT   2
#define SSRHT   3
#define SPKREN  4
#define KSWL    5
#define KSWR    6

// DAC Output channels
#define RIGHT_HI    0
#define LEFT_LO     1
#define RIGHT_LO    2
#define LEFT_HI     3

//-------------------------------------------------------------------------
// Direction controls (driver macros) - bits high for output
//-------------------------------------------------------------------------

#define INIT_DDR_B (BV(PB0) | BV(PB1) | BV(PB2) | BV(PB3) | BV(PB5))    // Incl. MOSI, SCK
#define INIT_DDR_C (BV(PC4) | BV(PC5))
#define INIT_DDR_D (BV(PD3) | BV(PD5))

//-Configuration constants ---------------------------------------------------
#define SIGNON      (PROJECT " " DESCRIPTION " v" REVISION ".  Copyright (c) " YEAR ". " AUTHOR "\n" \
                     "Built " __DATE__ "  " __TIME__ "\n")

#define ACCUMSHIFT  16              // Bit-shift for bias-servo accumulators

//-Symbols -------------------------------------------------------------------

typedef enum        // Panel-switch & remote input state 
{
    SW_LOCOFF = 0,
    SW_REMOFF,
    SW_STANDBY,
    SW_RUN,
    SW_ITEMS		// Should be last
} T_SWITCH;

typedef enum        // Main state machine
{
    M_FAULT,
    M_OFF,
    M_LTRAMP,
    M_WARMUP,
    M_STNDBY,
    M_HTRAMP,
    M_HTSTDY,
    M_CALIB,
    M_MUTE,
    M_V2ON,
    M_LIVE,
    M_RUN
} T_STATE;

typedef enum {CK_NONE, CK_OUT, CK_HIGH, CK_MOST, CK_ALL} CKVAL;  // What values to test
typedef enum{L_OFF,    L_SLOW, L_FAST, L_ON} 	LedState;
typedef enum{P_OFF,    P_ON,   P_RAMP} 	  		PwrState;
typedef enum{S_OFF,    S_ON}               		Switch;

typedef enum{LEFT=0, RIGHT} SIDES;
typedef enum{LOW=0,  HIGH}	BOUNDS;

typedef struct              // State machine table entry
{
    PwrState    LT, HT;			// Power-supply states
    Switch      V2;   			// V2 cathode drive
    Switch      Spkr;   		// Speaker relay
    pf_b_b      statefn;        // State actions: return controls flow
    CKVAL       LoBound;		// Test low voltage bounds
    T_STATE     succ[SW_ITEMS]; // Successor state, by control switch
    PidState    pid;			// Grid control action
    LedState    red, green;     // Panel LEDs
    uint1       name[8];        // Name for logger
} T_MAINST;

#define _WDR() __asm__ __volatile__ ("wdr" ::)

//-Global Externs ------------------------------------------------------------
extern const uint1 Fill[] flash;

extern T_SWITCH panelsw;
extern bool1    faultdet;
extern T_MAINST cstate;
extern uint2    seconds;
extern PGM_P    FaultMsg;
extern PGM_P    SwitchMsg;
extern bool1    HalfSec;        // Half-second prescaler
extern uint2    IV1[2];         // V1 cathode current	

//-Global Function Prototypes ------------------------------------------------
extern uint1 *GetStateName(void);
extern bool1  InFault(void);
extern void   ScreenSetup(bool1 calok);
extern void   SetWatchdog(void);

#endif /* __MAIN_H */

