//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file ancomp.c
//!
//! \project Igor-IIL 25+25W OTL valve amplifier
//!
//! \brief  Analogue comparator handling (ie AC powerline)
//!
//! The analogue comparator interrupts on both zero crossings of the AC
//! power line. This drives the main state machine.
//! This file includes handlers for the 8-bit timer T0, which is used
//! with the comparator, to generate PWM signals for the solid-state relays.
//!
//! \author David R Brooks
//!
//! \version 18/10/2007 DRB initial design and coding
//!
//---------------------------------------------------------------------------


//-Included files -----------------------------------------------------------
#include "main.h"

//-Local Symbols ------------------------------------------------------------
// [insert local #defines and variable types]

//-Globals ------------------------------------------------------------------
uint2 PowerCycle = 0;			// TC0 counts per AC cycle (128 or 160)
uint1 SsrMask    = ~(BV(SSRLT) | BV(SSRHT)); // Short-pulse the SSR outputs
uint1 PulseDelay = 0;		    // SSR pulse delay

//-Locals -------------------------------------------------------------------
static uint1 SsrSet     = 0;		// Set ON the SSR bits
static uint1 AcState    = 0;		// Measure the AC-line period

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

static inline bool1 VarPhase(uint1 x) {return((x > 0) && (x < 255));};
static void   FixPhase(uint1 ssrbit, uint1 val);

//! \endcond

//-Local Functions ----------------------------------------------------------
// [insert local functions]

//-Global Functions ---------------------------------------------------------

//--IsAcLineReady--------------------------------------------------------------
//! Check if the AC frequency measurement is complete
//-----------------------------------------------------------------------------
//! 
//! \return           TRUE if done
//!
//! \note             Can't be static/inline, due to limited scope of AcState
//! 
//-----------------------------------------------------------------------------
bool1 IsAcLineReady(void)
{
	return (AcState == 13);
}

//--SIGNAL---------------------------------------------------------------------
//! Analogue comparator interrupt function
//-----------------------------------------------------------------------------
//! Interrupts on both edges of the comparator, ie every half-cycle of the
//! AC line. Generates phase controls for the solid-state relays.
//!
//! The PWM width is given in units of 125uS (8kHz), & is loaded as a 2's complement.
//! 
//-----------------------------------------------------------------------------
SIGNAL(SIG_COMPARATOR)
{
	PORTD ^= BV(PD5);    
    switch(AcState)
    {
    case 10:						// Start measurement after 5 cycles warmup
    	TCNT0 = 0;
	    TCCR0 = 5;					// 8kHz clock
    	break;
    case 12:						// Full cycle: complete measurements
    	PowerCycle = TCNT0;
        if((TCNT0 < 100) || (TCNT0 > 180))	// Failed to get a good reading
        {
        	AcState = 0;					// Restart the measurement
        }
        else
        {
	        TCNT0  = 0;
	        TCCR0  = 0;				// T0 stopped until requested
	    	TIMSK |= 1;				// T0 interrupt on now
        }
    	break;
    case 13:						// Normal running
    	if(PulseDelay != 0)
        {
			TCNT0 = PulseDelay;		// Start the monostable timer
		    TCCR0 = 5;				// 8kHz clock
        }
        return;						// Don't increment AcState
    default:
    	break;						// Just count forward
    }
    AcState++;
}

//--SIGNAL---------------------------------------------------------------------
//! T0 overflow interrupt
//-----------------------------------------------------------------------------
//! T0 is triggered by an analogue-comparator interrupt (ie AC line zero crossing).
//! It acts as a monostable, to generate a PWM signal synchronised to the AC line.
//! This drives the solid-state relays for the main power.
//! 
//-----------------------------------------------------------------------------
SIGNAL(SIG_OVERFLOW0)
{
	Reg595 |= SsrSet;			// Set on any SSR starts

	TCCR0 = 0;					// Stop T0 (ie monostable action)
}

//--SetPhase-------------------------------------------------------------------
//! Set the conduction angles for the AC lines
//-----------------------------------------------------------------------------
//! Angle is expressed as a fraction of 256 (0=off, 255=full-on). Only one
//! circuit may have a partial phase at any one time: the other must be 0 or 255.
//! 
//! \param HtAngle      Phase for HT circuit
//! \param LtAngle      Phase for LT circuit
//!
//! \return           TRUE if successful
//! 
//-----------------------------------------------------------------------------
bool1 SetPhase(uint1 HtAngle, uint1 LtAngle)
{
	uint1 mode = 0;
    if(VarPhase(LtAngle))
    {
    	mode |= 1;
    }
    if(VarPhase(HtAngle))
    {
    	mode |= 2;
    }
	switch(mode)
    {
    case 1:				// Set variable angle
    	mode = 255 - LtAngle;
        break;
    case 2:
    	mode = 255 - HtAngle;
        break;
    case 0:				// Let angle be zero (not used)
    	break;
    default:
    	return FALSE;
    }
    
    PulseDelay = -(PowerCycle * mode / 512);

	FixPhase(BV(SSRHT), HtAngle);
	FixPhase(BV(SSRLT), LtAngle);
    return TRUE;
}


//--FixPhase-------------------------------------------------------------------
//! Set bit-masks to drive the SSR lines
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param ssrbit     Bit-mask to set
//! \param val        Phase value (0..255)
//! 
//-----------------------------------------------------------------------------
static void FixPhase(uint1 ssrbit, uint1 val)
{
	switch(val)
    {
    case 0:				// Turn off
    	SsrMask  &= ~ssrbit;		// Stop the output
        SsrSet   &= ~ssrbit;
    	break;
    case 255:			// Turn on
    	SsrMask  |= ssrbit;			// Hold the output on
        SsrSet   |= ssrbit;
        Reg595   |= ssrbit;
    	break;
    default:			// Set phase
    	SsrMask  &= ~ssrbit;		// Pulse the output
        SsrSet   |= ssrbit;			// On at start
    }
}
