//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file serial.c
//!
//! \project Igor-IIL 25+25W OTL valve amplifier
//!
//! \brief  Asynchronous serial comms. handler
//!
//! Interrupt driven serial TX
//!     Version:    1.01         
//!     Processor:  Atmel ATMega8
//!     Compiler:   GCC          
//!
//! \author David R Brooks
//!
//! \version 13/07/2007 DRB initial design and coding
//!
//! \todo                Receiver functions
//---------------------------------------------------------------------------
// 2008.07.28 18:47  [YYYY.MM.DD] Last edited (CodeWright)

//-Included files -----------------------------------------------------------
#include "main.h"
#include <stdlib.h>

/*
Teraterm cursor & function keys (VT100):
UP:   1B 5B 41        E[A
DN:   1B 5B 42        E[B
RT:   1B 5B 43        E[C
LT:   1B 5B 44        E[D

HOME: 1B 5B 32 7E     E[2~
PGUP: 1B 5B 33 7E     E[3~
END:  1B 5B 35 7E     E[5~   NB Breaks seq.
PGDN: 1B 5B 36 7E     E[6~

F1:   1B 5B 31 31 7E  E[11~
F2:   1B 5B 31 32 7E  E[12~
F3:   1B 5B 31 33 7E  E[13~
F4:   1B 5B 31 34 7E  E[14~
F5:   1B 5B 31 35 7E  E[15~
F6:   1B 5B 31 37 7E  E[17~   NB Breaks seq.
F7:   1B 5B 31 38 7E  E[18~
F8:   1B 5B 31 39 7E  E[19~
F9:   1B 5B 32 30 7E  E[20~
F10:  1B 5B 32 31 7E  E[21~
F11:  1B 5B 32 33 7E  E[23~   NB Breaks seq.
F12:  1B 5B 32 34 7E  E[24~
*/

//-Local Symbols ------------------------------------------------------------
#define CSI	"\033" "["					// Don't put brackets round: merge strings

typedef struct
{
    uint2   speed_u2;
    uint2   factor_u2;
} T_COMMS;    

//-Globals ------------------------------------------------------------------
// [insert global variables]

//-Locals -------------------------------------------------------------------
static uint1 TXQ[32];
static uint1 TQpush;
static uint1 TQpull;
static bool1 TQactive;

static const T_COMMS BaudTab[] flash = {   // Baud-rate table (8MHz oscillator, double speed)
    {  2400, 416 },
    {  4800, 207 },
    {  9600, 103 },
    { 14400,  68 },
    { 19200,  51 },
    { 38400,  25 },
    { 57600,  16 }			// Highest possible in 2 bytes. Last entry is default.
};

static uint1 RxChar = '\0';

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

static bool1 IsTQempty(uint1* ppt);
static void  ToTQ(uint1 c);
static uint1 FmTQ(void);
static bool1 TxChar(void);

//! \endcond

//-Local Functions ----------------------------------------------------------

//--TX interrupt-------------------------------------------------------------
//! Serial TX interrupt
//---------------------------------------------------------------------------
//! This interrupt occurs when the TX buffer becomes free, ie a char. has been
//! moved to the shift register.
//! 
//---------------------------------------------------------------------------
SIGNAL(SIG_UART_TRANS)
{
	sei;
	(void)TxChar();
}

//--TX sender----------------------------------------------------------------
//! Pull a char. from the queue, and send.
//---------------------------------------------------------------------------
//! This is called both from the TX interrupt, and from the FIFO driver (to
//! start a new transmission).
//!
//! \return      TRUE if a char. was sent, FALSE if queue was empty
//!
//! \note        Run with ints. disabled
//! 
//---------------------------------------------------------------------------
static bool1 TxChar(void)
{
	uint1 c = FmTQ();			// Next char. from main FIFO
	
	if(c)
	{
		UDR = c;				// Send to TX
		return TRUE;
	}
	else
	{
		TQactive = FALSE;		// Nothing to send
		return FALSE;
	}
}

//--SIGNAL---------------------------------------------------------------------
//! Serial RX interrupt
//-----------------------------------------------------------------------------
//! Accepts characters, & assembles ANSI command sequences.
//! ANSI sequences must be handled here, as the outer fns. are called by the 
//! base-level poll loop, & so will not keep up (at high baud rates) with the line.
//! 
//-----------------------------------------------------------------------------
SIGNAL(SIG_UART_RECV)
{
    //                           2	    3	   4  5	    6	   7  8  9  10 11   12   13   14   15   16 17   18   19   20   21    22 23    24
	static const uint1 code[] = {KHOME, KPGUP, 0, KEND, KPGDN, 0, 0, 0, 0, KF1, KF2, KF3, KF4, KF5, 0, KF6, KF7, KF8, KF9, KF10, 0, KF11, KF12};
    static const uint1 move[] = {KUP, KDOWN, KRIGHT, KLEFT};
	static uint1 arg;
	static uint1 state = 0;
	uint1            c = UDR;
	sei;
    
    switch(state)
    {
    case 0:					// Nothing special
    default:
    	switch(c)
        {
        case '\033':
        	state = 1;
            arg   = 0;
            return;
        default:
        	;
        }
        break;
        
    case 1:					// ESC seen...
    	switch(c)
        {
        case '[':
        	state = 2;
            return;
        default:
        	;
        }
        break;
    case 2:					// ESC [ seen...
		state = 0;				// End of most ESC sequences
    	switch(c)
        {
        case '0' ... '9':		// Numeric argument
        	arg = (arg * 10) + (c - '0');
            state = 2;			// Keep this state
            return;
        case '~':				// Executive for ESC [ nn ~
        	switch(arg)
            {
            case 2 ... 24:		// Function keys
            	RxChar = code[arg - 2];
                return;
            default:
                ;
            }
            break;
        case 'A' ... 'D':		// Cursor move keys
        	RxChar = move[c - 'A'];
         	return;
        default:
    		;
        }
    }
	state  = 0;				// No ESC sequence here
	RxChar = c;
}

//--SerialGetRx--------------------------------------------------------------
//! Return polled keyboard char. (or \0)
//---------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \return           Returned char.
//! 
//---------------------------------------------------------------------------
uint1 SerialGetRx(void)
{
	uint1 c;
    
    cli;
    c = RxChar;
    RxChar = '\0';
    sei;
    return c;
}

//--IsTQempty----------------------------------------------------------------
//! Return TRUE if the FIFO is/will become empty
//---------------------------------------------------------------------------
//!
//! \param ppt             Addr. of value to use as push pointer
//! \return                TRUE if empty
//! 
//---------------------------------------------------------------------------
static bool1 IsTQempty(uint1* ppt)
{
	if(ppt)
	{
		return(*ppt == TQpull);
	}
	else
	{
		return(TQpush == TQpull);
	}
}

//--TQsucc-------------------------------------------------------------------
//! TX FIFO pointer's successor function
//---------------------------------------------------------------------------
//! Get next FIFO locn.
//! 
//! \param ptr          Current ptr.
//!
//! \note				sizeof(TXQ) must be a power of 2
//! 
//---------------------------------------------------------------------------
static inline uint1 TQsucc(uint1 ptr)
{
	ptr++;
	ptr &= (sizeof(TXQ)-1);
	return ptr;
}

//--ToTQ---------------------------------------------------------------------
//! Put a char. in the queue
//---------------------------------------------------------------------------
//! If the queue is full, wait.
//! 
//! \param c              Char. to load
//!
//---------------------------------------------------------------------------
static void ToTQ(uint1 c)
{
	uint1 pt = TQsucc(TQpush);
	
	while(IsTQempty(&pt))			// Wait, if no room in the queue?
		;

	TXQ[TQpush] = c;				// Load the char.
	TQpush = pt;
}

//--FmTQ---------------------------------------------------------------------
//! Get a char. from the FIFO, if any
//---------------------------------------------------------------------------
//! If the queue is empty, return NUL
//! 
//! \return               Returned char, or NUL
//!
//---------------------------------------------------------------------------
static uint1 FmTQ(void)
{
	if(IsTQempty(NULL))
	{
		return 0;			// FIFO is empty
	}
	else
	{
		uint1 c = TXQ[TQpull];
		TQpull  = TQsucc(TQpull);
		return c;
	}
}

//-Global Functions ---------------------------------------------------------

//--CommsInit----------------------------------------------------------------
//! Initialise UART (async. use)
//---------------------------------------------------------------------------
//! Start the comms. port, for TX interrupt, RX polled
//! Given an invalid rate selection, the last table entry is used.
//! 
//! \note                 Assumes 8MHz clock, and double-speed UART
//!
//---------------------------------------------------------------------------
void CommsInit(uint2 rate_u2)
{
    int     i;
	T_COMMS tab;

    for(i=0; i<DIMOF(BaudTab); i++)     // Table lookup
    {
        memcpy_P(&tab, &BaudTab[i], sizeof(tab));
        if(rate_u2 == tab.speed_u2)		// Found the reqd. speed
        {
        	break;						// If we don't match, last entry is default
        }
    }
    
    UDR   = 0;                                  			// Clear data register
    UCSRA = BV(U2X);                            			// Double clock speed
    UCSRB = BV(RXEN)  | BV(TXEN)  | BV(TXCIE) | BV(RXCIE);  // RX & TX on (full duplex)
    UCSRC = BV(URSEL) | BV(UCSZ1) | BV(UCSZ0);  			// Mode: 8N1
    UBRRH = tab.factor_u2 >> 8;                 			// Baud rate: 16 bits
    UBRRL = tab.factor_u2 & 0x00ff;
}

//--outn-----------------------------------------------------------------------
//! Output n as an unsigned string, radix r
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param value      Value to output
//! \param width      Width to display
//! \param radix      Radix
//! \param sign       TRUE if signed
//! 
//-----------------------------------------------------------------------------
void outn(uint2 value, sint1 width, sint1 radix, bool1 sign)
{
	static const uint1 t[] flash = "0123456789ABCDEF";

    if(sign && (value & 0x8000))
    {
    	uint2 pval = (uint2)(-((sint2)value));
        outn(pval, 0x40 | (width-1), radix, FALSE);
    }
    else
    {
    	div_t qr;
        		
        if(value & 0x8000)						// div() assumes signed
        {
        	uint2 t = value & 1;				// Save the LS bit
            value   = (value >> 1) & 0x7FFF;	// Pre-scale the division
			qr      = div(value, radix / 2); 	// Quotient is correct
            qr.rem  = (qr.rem << 1) + t;		// Correct the remainder
        }
        else
        {										// value is in-range for div()
			qr = div(value, radix); 			// Quotient & remainder
        }
        
        if(qr.quot > 0)
        {
            outn(qr.quot, width-1, radix, FALSE);
        }
        else
        {
        	uint1 neg = width & 0x40;
            width &= 0x3f;					// Remove the sign tag
			for(; width>0; width--)
			{
				outc(' ');
			}
            if(neg)
            {
            	outc('-');
            }
        }
        outc(pgm_read_byte(&t[qr.rem]));
    }
}

//--outc---------------------------------------------------------------------
//! Basic char. output on UART
//---------------------------------------------------------------------------
//! Interrupt driven
//! 
//! \param c          Char. to send
//! 
//! \pre              Assumes port is already set up (see CommsPortInit)
//! \note             LF is expanded to CR/LF
//! \sa               CommsPortInit()
//! 
//---------------------------------------------------------------------------
void outc(char c)
{
    if(c == '\n')                   // LF -> CR-LF
        outc('\r');
		
	ToTQ(c);						// Waits if queue is full

	cli;                            // Disable, in case the FIFO was emptied just here
	if(!TQactive)					// Need to start it?
	{
		TQactive = TxChar();		// Try to start the TX
	}
	sei;
}

//--outsw----------------------------------------------------------------------
//! Output string, pad to width
//-----------------------------------------------------------------------------
//! [insert detailed description]
//!
//! \param p          Ptr. to string
//! \param w          Minimum width
//!
//-----------------------------------------------------------------------------
void outsw(char* p, sint2 w)
{
    for(w -= outs(p); w > 0; w--)
    {
        outc(' ');
    }
}

//--outsw_P--------------------------------------------------------------------
//! Output string, pad to width (from flash)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//!
//! \param p          Ptr. to string
//! \param w          Minimum width
//!
//-----------------------------------------------------------------------------
void  outsw_P(PGM_P p, sint2 w)
{
    for(w -= outs_P(p); w > 0; w--)
    {
        outc(' ');
    }
}

//--outs---------------------------------------------------------------------
//! Output a string, return its width on screen
//---------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param p          Ptr. to string
//! \return           Displayed width of string
//! 
//---------------------------------------------------------------------------
uint1 outs(char* p)
{
    uint1 l = 0;

	while(*p)
	{
        if(*p >= ' ')
        {                   // Count of displayable chars.
            l++;
        }
		outc(*p++);
	}
    return l;
}

//--outs_P---------------------------------------------------------------------
//! Output a string from code flash
//-----------------------------------------------------------------------------
//! Similar to outs(), but accepts a pointer into code flash, not RAM.
//! Uses the std. library to access code flash as data.
//! 
//! \param p          String ptr. (in code flash)
//! \return           Displayed width of string
//! 
//! \note             Strings must be defined using prog_char
//-----------------------------------------------------------------------------
uint1 outs_P(PGM_P p)
{
    char c;
    uint1 l = 0;

	while((c = pgm_read_byte(p++)))
	{
        if(c >= ' ')
        {                   // Count of displayable chars.
            l++;
        }
		outc(c);
   		_WDR();
	}
    return l;
}

//--wait---------------------------------------------------------------------
//! Simple NuS delay
//---------------------------------------------------------------------------
//! A simple delay loop, about 1uS per count
//! 
//! \param delay      Delay in uS
//! 
//---------------------------------------------------------------------------
void wait(uint4 delay_us)
{
    while(delay_us--)
	{
        __asm__ __volatile__ ("nop" ::);	// Keeps GCC from optimising it away
	}
}

//--IsTxqEmpty---------------------------------------------------------------
//! Check if the TX FIFO is empty
//---------------------------------------------------------------------------
//! 
//! \return        TRUE is it is empty
//!
//---------------------------------------------------------------------------
bool1 IsTxqEmpty(void)
{
	return IsTQempty(NULL);
}

/***************************************************************************
           S U P P O R T   F O R   V T - 1 0 0   T E R M I N A L
 ***************************************************************************/

//--ClearScreen----------------------------------------------------------------
//! Erase the whole screen: cursor home
//-----------------------------------------------------------------------------
//! 
//-----------------------------------------------------------------------------
void ClearScreen(void)					// ED; CUP
{
	outs_P(PSTR(CSI "2J"));				// Erase
    SetCursor(1,1);						// Home
}

//--SetCursor------------------------------------------------------------------
//! Move cursor to given posn. (Top left is (1,1)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param row        Target row (based 1)
//! \param col        Target col. (based 1)
//! 
//-----------------------------------------------------------------------------
void SetCursor(uint1 row, uint1 col)	// CUP
{
	outs_P(PSTR(CSI));					// Set cursor posn.
    PutByte(row);
	outc(';');
    PutByte(col);
	outc('H');
}

//--PutByte--------------------------------------------------------------------
//! Output a single value, 0..99
//-----------------------------------------------------------------------------
//! Used typically by SetCursor, etc.
//! No signs, leading space/zeros, etc.
//! 
//! \param x          Value to send
//! 
//-----------------------------------------------------------------------------
void PutByte(uint1 x)
{
	if(x > 9)
    {
		PutByte(x / 10);
    }
    outc('0' + (x % 10));
}

//--DeleteLine-----------------------------------------------------------------
//! Delete cursor line: scroll up from below
//-----------------------------------------------------------------------------
//! NB This is supported by TeraTerm, but not a generic VT100
//! 
//! \param count      No. of lines to delete
//! 
//-----------------------------------------------------------------------------
void DeleteLine(uint1 count)			// DL (Teraterm 4, not VT100)
{
	outs_P(PSTR(CSI));					// Erase current line/s, scroll up below
	outn(count, 0, 10, FALSE);
    outc('M');
}
