//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file serial.h
//!
//! \project Igor-IIL 25+25W OTL valve amplifier
//!
//! \brief  Publicly accessible functions and types for serial.c
//!
//! Asynchronous serial comms. handler
//!     Version:    1.01         
//!     Processor:  Atmel ATMega8
//!     Compiler:   GCC          
//!
//! \author David R Brooks
//! \date 13/07/2007
//!
//---------------------------------------------------------------------------
// 2008.06.07 12:36  [YYYY.MM.DD] Last edited (CodeWright)

#ifndef __SERIAL_H
#define __SERIAL_H

//-Symbols -------------------------------------------------------------------
#include "typedefn.h"

#define SCREENWIDTH		80
#define SCREENHEIGHT	24

typedef enum 			// 1-key codes for ANSI functions
{
	KUP		= 0xA0,
    KDOWN	= 0xA1,
    KLEFT	= 0xA2,
    KRIGHT	= 0xA3,
    KHOME	= 0xA4,
    KEND	= 0x83,
    KPGUP	= 0x85,
    KPGDN	= 0x86,
    KF1		= 0x8B,
    KF2		= 0x8C,
    KF3		= 0x8D,
    KF4		= 0x8E,
    KF5		= 0x8F,
    KF6		= 0x91,
    KF7		= 0x92,
    KF8		= 0x93,
    KF9		= 0x94,
    KF10	= 0x95,
    KF11 	= 0x97,
    KF12	= 0x98,
}C_KEY;

//-Global Externs ------------------------------------------------------------

//-Global Function Prototypes ------------------------------------------------

//--CommsInit----------------------------------------------------------------
//! Initialise UART (async. use)
//---------------------------------------------------------------------------
//! Start the comms. port, for TX interrupt, RX polled
//! 
//! \param rate_u4        Speed setting (bits/sec)
//!
//! \note                 Assumes 8MHz clock, and double-speed UART
//!
//---------------------------------------------------------------------------
void CommsInit(uint2 rate_u2); 	// Start the comms. port, for TX interrupt, RX polled      

//--outc---------------------------------------------------------------------
//! Basic char. output on UART
//---------------------------------------------------------------------------
//! Interrupt driven
//! 
//! \param c          Char. to send
//! 
//! \pre              Assumes port is already set up (see CommsPortInit)
//! \note             LF is expanded to CR/LF
//! \sa               CommsPortInit()
//! 
//---------------------------------------------------------------------------
void  outc(char c);             	// Basic char. output on UART

//--outs---------------------------------------------------------------------
//! Output a string, return its width on screen
//---------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param p          Ptr. to string
//! \return           Displayed width of string
//! 
//---------------------------------------------------------------------------
uint1 outs(char* p);            	// Output a string, return its width on screen

//--outs_P---------------------------------------------------------------------
//! Output a string from code flash
//-----------------------------------------------------------------------------
//! Similar to outs(), but accepts a pointer into code flash, not RAM.
//! Uses the std. library to access code flash as data.
//! 
//! \param p          String ptr. (in code flash)
//! \return           Displayed width of string
//! 
//! \note             Strings must be defined using prog_char
//-----------------------------------------------------------------------------
uint1 outs_P(PGM_P p);          	// Output a string from code flash

//--outsw----------------------------------------------------------------------
//! Output string, pad to width
//-----------------------------------------------------------------------------
//! [insert detailed description]
//!
//! \param p          Ptr. to string
//! \param w          Minimum width
//!
//-----------------------------------------------------------------------------
void  outsw(char* p, sint2 w);		// Output string, pad to width

//--outsw_P--------------------------------------------------------------------
//! Output string, pad to width (from flash)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//!
//! \param p          Ptr. to string
//! \param w          Minimum width
//!
//-----------------------------------------------------------------------------
void  outsw_P(PGM_P p, sint2 w);	// Output string from Flash, pad to width

//--outn-----------------------------------------------------------------------
//! Output n as an unsigned string, radix r
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param value      Value to output
//! \param width      Width to display
//! \param radix      Radix
//! \param sign       TRUE if a signed value
//! 
//-----------------------------------------------------------------------------
void  outn(uint2 value, sint1 width, sint1 radix, bool1 sign);	// Output value in width chars, in radix

//--wait---------------------------------------------------------------------
//! Simple NuS delay
//---------------------------------------------------------------------------
//! A simple delay loop, about 1uS per count
//! 
//! \param delay      Delay in uS
//! 
//---------------------------------------------------------------------------
void  wait(uint4 delay_us);     	// A simple delay loop, about 1uS per count

//--IsTxqEmpty---------------------------------------------------------------
//! Check if the TX FIFO is empty
//---------------------------------------------------------------------------
//! 
//! \return        TRUE is it is empty
//!
//---------------------------------------------------------------------------
bool1 IsTxqEmpty(void);         	// Check if the TX FIFO is empty

//--SerialGetRx----------------------------------------------------------------
//! Return polled keyboard char. (or \0)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \return           Returned char.
//! 
//-----------------------------------------------------------------------------
uint1 SerialGetRx(void);        	// Get RX char. or NUL

// ========== VT-100 terminal support ======================

//--ClearScreen----------------------------------------------------------------
//! Erase the whole screen: cursor home
//-----------------------------------------------------------------------------
//! 
//-----------------------------------------------------------------------------
void  ClearScreen(void);				// Erase everything, cursor home

//--SetCursor------------------------------------------------------------------
//! Move cursor to given posn. (Top left is (1,1)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param row        Target row (based 1)
//! \param col        Target col. (based 1)
//! 
//-----------------------------------------------------------------------------
void  SetCursor(uint1 row, uint1 col);	// Cursor to row & col (based 1)

//--DeleteLine-----------------------------------------------------------------
//! Delete cursor line: scroll up from below
//-----------------------------------------------------------------------------
//! NB This is supported by TeraTerm, but not a generic VT100
//! 
//! \param count      No. of lines to delete
//! 
//-----------------------------------------------------------------------------
void  DeleteLine(uint1 count);			// Delete current line/s, scroll up below

void PutByte(uint1 x);

#endif /* __SERIAL_H */
