//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file spi.c
//!
//! \project Igor-IIL 25+25W OTL valve amplifier
//!
//! \brief  Support for the SPI driver (DAC & output register)
//!
//! SPI output driver. This serves the following outputs:
//! 1. 4-channel DAC feeding the output tube bias generators
//! 2. 74HC595 SIPO register (relay drivers, etc.)
//! Optionally, it may also feed the following (if used):
//! 3. Realtime clock
//! 4. Flash ROM
//!
//! Currently, only 1 & 2 are used. Operation is autonomous: the higher
//! levels simply update a RAM image, which is autonomously copied to the
//! SPI lines.
//!
//! \author David R Brooks
//!
//! \version 27/10/2007 DRB initial design and coding
//!
//---------------------------------------------------------------------------
// 2008.08.16 16:21  [YYYY.MM.DD] Last edited (CodeWright)


//-Included files -----------------------------------------------------------
#include "main.h"

//-Local Symbols ------------------------------------------------------------
// [insert local #defines and variable types]

//-Globals ------------------------------------------------------------------
								// NB Update DacValue atomically!
uint2 DacValue[4] = {248,540,540,248};		// Values to send to TLV5604 DAC
uint1 Reg595      = (BV(KSWL) | BV(KSWR));	// Value  to send to 74HC595 output bits
sint2 DacDelta[4] = {0};

//-Locals -------------------------------------------------------------------
// See spi.h for SpiInit()

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

// [insert local protoypes]

//! \endcond

//-Local Functions ----------------------------------------------------------
// [insert local functions]

//-Global Functions ---------------------------------------------------------

//--DacUpdate------------------------------------------------------------------
//! Update the DAC drive values from the errors
//-----------------------------------------------------------------------------
//! For each OP grid, compute & update the bias value.
//! This is the core of the whole project: maintain correct bias on the
//! OP valves. It's a very simple PI-like action.
//! The action DOES NOT invert: ie a too-high measurement will raise the output.
//! This is correct, as the output drivers provide inversion, 
//! DAC=0, Vo=high: DAC=5, Vo=low
//! 
//-----------------------------------------------------------------------------
void DacUpdate(PidState fast)
{								//      RU      LL      RL     LU
	static const E_ADC_CHAN ichn[4] = {VOUTR, ICLIPL, ICLIPR, VOUTL};	// Input ADC channels
	uint1 i;
    uint2 scale = 1;
    sint4 diff;
    for(i=0; i<4; i++)
    {
        switch(fast)
        {
        case G_OFF:
            DacSetValue(i, 0x3FF);		// Max. value, ie full negative grids
            continue;
        case G_WIDE:
            scale = CalRam.fastscale;
            break;
        case G_NARROW:
            scale = CalRam.servo[i].dacscale;
            break;
        }
    //                     Measurement            Setpoint
    	DacDelta[i] = ScaledAdc_s2[ichn[i]] - CalRam.servo[i].setpoint;
        diff = DacDelta[i];			// Be sure it sign-extends properly
    	CalRam.servo[i].dacdrive += scale * diff;
        DacSetValue(i, CalRam.servo[i].dacdrive >> ACCUMSHIFT);		// Top 10 bits > DAC
    }	
}

//--SIGNAL---------------------------------------------------------------------
//! SPI interrupt function
//-----------------------------------------------------------------------------
//! The SPI runs continuously, updating the DAC and output latches
//! Software strategies need only update the buffer values.
//! Update period (3 bytes) = 1.5mS
//! 
//! \note             DAC word: CH1 CH0 PDWN FAST D[9:0] X X 
//!
//-----------------------------------------------------------------------------
SIGNAL(SIG_SPI)
{
	static uint1 state = 0;
    static uint1 chan  = 0;
    static uint2 cdac;			// Atomic snapshot for DAC channel

	sei;
    switch(state)
    {
    default:					// Wrap, or force it back into sync.
    	state = 0;
        // FALL THROUGH
    case 0:						// Send first (MS) byte to DAC
    	cli;
        cdac = DacValue[chan];	// Atomic snapshot
        sei;
    	SELLATCH(1);
    	DACSET(0);
        SETBIT(SPCR, CPOL, 1);
    	SELLATCH(0);			// Latch the just-sent output-bits data
        cdac = (chan << 14) | 0x1000 | ((cdac << 2) & 0x0FFC);   // Create DAC control word	
        SPDR = cdac >> 8; 
    	break;
    case 1:						// Send second (LS) byte to DAC
    	SPDR = cdac & 0x00ff;
        chan = (chan + 1) & 3;
    	break;					
    case 2:						// Send update to output-pins latch
    	DACSET(1);				// Latch the DAC data
        SETBIT(SPCR, CPOL, 0);
        if(faultdet)			// Un-processed fault stops everything
        {
        	SPDR = 0;				// All relays (& LEDs) off
        }
        else
        {
        	SPDR = Reg595;
        	Reg595 &= SsrMask;	// Short-pulse the SSR outputs
        }
    	break;
    }
    state++;
}

//--DacSetValue----------------------------------------------------------------
//! Set new value on given DAC channel
//-----------------------------------------------------------------------------
//! The value will update on the next loop of the SPI engine.
//! 
//! \param chan       Channel, 0-3
//! \param val        Value 0-1023
//! 
//! \warning          Run from base level: interrupts are re-enabled
//!
//-----------------------------------------------------------------------------
void DacSetValue(uint1 chan, uint2 val)
{
    chan &= 3;
    val  &= 0x3FF;
    
    cli;						// Update 2-byte DAC register, safely
	DacValue[chan] = val;
    sei;
}

